
import numpy as np
import matplotlib.pyplot as plt
from .image import Image
from .camera import Camera

#%% LevelSet Calibration tools

def LSfromLine(f, pts1):
    """Compute level set of a line from a point cloud"""
    b = pts1.T.dot(np.ones_like(pts1[:, 0]))
    A = pts1.T.dot(pts1)
    res = np.linalg.solve(A, b)
    ui = np.arange(0, f.pix.shape[0])
    vi = np.arange(0, f.pix.shape[1])
    [Yi, Xi] = np.meshgrid(vi, ui)
    lvlset = (Xi * res[1] + Yi * res[0] - 1) / np.linalg.norm(res)
    lvl = Image("lvl")
    lvl.pix = abs(lvlset)
    return lvl

def LSfromPoint(f, pts1):
    """Compute level set from one single point"""
    pts1 = pts1.ravel()
    ui = np.arange(f.pix.shape[0]) - pts1[1]
    vi = np.arange(f.pix.shape[1]) - pts1[0]
    [Yi, Xi] = np.meshgrid(vi, ui)
    lvl = Image("lvl")
    lvl.pix = np.sqrt(Xi ** 2 + Yi ** 2)
    return lvl

def LSfromCircle(f, pts1):
    """Compute level set of a circle from a point cloud"""
    meanu = np.mean(pts1, axis=0)
    pts = pts1 - meanu
    pts2 = pts ** 2
    A = pts.T.dot(pts)
    b = 0.5 * np.sum(pts.T.dot(pts2), axis=1)
    cpos = np.linalg.solve(A, b)
    R = np.sqrt(np.linalg.norm(cpos) ** 2 + np.sum(pts2) / pts.shape[0])
    cpos += meanu
    ui = np.arange(0, f.pix.shape[0])
    vi = np.arange(0, f.pix.shape[1])
    [Yi, Xi] = np.meshgrid(vi, ui)
    lvlset = abs(np.sqrt((Xi - cpos[1]) ** 2 + (Yi - cpos[0]) ** 2) - R)
    lvl = Image("lvl")
    lvl.pix = abs(lvlset)
    return lvl

class LSCalibrator:
    """Calibration of a front parallel setting 2D-DIC"""
    def __init__(self, f, m):
        self.f = f
        self.m = m
        self.ptsi = dict()
        self.ptsm = dict()
        self.feat = dict()
        self.nfeat = 0
        self.lvl = dict()
        self.cam = None

    def Init3Pts(self, ptsm=None, ptsM=None):
        """Initialization of the calibration using 3 points.

        Parameters
        ----------
        ptsm : Numpy array
            points coordinates in the images (DEFAULT = defined by clic)
        ptsM : Numpy array
            points coordinates in the mesh (DEFAULT = defined by clic)
            
        """
        if ptsm is None:
            print(" ************************************************* ")
            print(" *  SELECT 3 characteristic points in the image  * ")
            print(" ************************************************* ")
            ptsm = self.f.SelectPoints(3)[:, [1, 0]]
        if ptsM is None:
            print(" ************************************************* ")
            print(" * SELECT the 3 corresponding points on the mesh * ")
            print(" ************************************************* ")
            ptsM = self.m.SelectPoints(3)
        cm = np.mean(ptsm, axis=0)
        cM = np.mean(ptsM, axis=0)
        dm = np.linalg.norm(ptsm - cm, axis=1)
        dM = np.linalg.norm(ptsM - cM, axis=1)
        scale = np.mean(dm / dM)
        dmax = np.argmax(dm)
        vm = ptsm[dmax] - cm
        vM = ptsM[dmax] - cM
        vm /= np.linalg.norm(vm)
        vM /= np.linalg.norm(vM)
        angl = np.arccos(vM @ vm)
        p = np.array([scale, 0, 0, np.pi / 2 - angl])
        self.cam = Camera(p)
        for i in range(40):
            up, vp = self.cam.P(ptsM[:, 0], ptsM[:, 1])
            dPudp, dPvdp = self.cam.dPdp(ptsM[:, 0], ptsM[:, 1])
            A = np.vstack((dPudp, dPvdp))
            M = A.T @ A
            b = A.T @ (ptsm.T.ravel() - np.append(up, vp))
            dp = np.linalg.solve(M, b)
            p += 0.8 * dp
            self.cam.set_p(p)
            err = np.linalg.norm(dp) / np.linalg.norm(p)
            print(
                "Iter # %2d | disc=%2.2f %% | dU/U=%1.2e"
                % (i + 1, np.linalg.norm(ptsm.T.ravel() - np.append(up, vp))
                    / np.linalg.norm(ptsm.T.ravel()) * 100, err))
            if err < 1e-5:
                break

    def Plot(self):
        """Plot the level sets of each feature"""
        for i in self.feat.keys():
            plt.figure()
            self.f.Plot()
            plt.contour(self.lvl[i].pix, np.array([0.4]), colors=["y"])
            plt.figure()
            plt.contourf(self.lvl[i].pix, 16, origin="image")
            plt.colorbar()
            plt.contour(self.lvl[i].pix, np.array([0.4]), colors=["y"], origin="image")
            plt.axis("image")

    def NewCircle(self):
        print(" ******************************* ")
        print(" *        SELECT Circle        * ")
        self.ptsi[self.nfeat] = self.f.SelectPoints(
            -1, title="Select n points of a circle... and press enter"
        )  # [:,[1,0]]
        self.ptsm[self.nfeat] = self.m.SelectCircle()
        self.feat[self.nfeat] = "circle"
        self.nfeat += 1
        print(" ******************************* ")

    def NewLine(self):
        print(" ******************************* ")
        print(" *        SELECT Line          *")
        self.ptsi[self.nfeat] = self.f.SelectPoints(
            -1, title="Select n points of a straight line... and press enter"
        )  # [:,[1,0]]
        self.ptsm[self.nfeat] = self.m.SelectLine()
        self.feat[self.nfeat] = "line"
        self.nfeat += 1
        print(" ******************************* ")

    def NewPoint(self):
        print(" ******************************* ")
        print(" *        SELECT Point         * ")
        self.ptsi[self.nfeat] = self.f.SelectPoints(1)  # [:,[1,0]]
        self.ptsm[self.nfeat] = self.m.SelectNodes(1)
        self.feat[self.nfeat] = "point"
        self.nfeat += 1
        print(" ******************************* ")

    def DisableFeature(self, i):
        """Disable one of the features. Used to redefine an inappropriate mesh selection.

        Parameters
        ----------
        i : int
            the feature number
            
        """
        if i in self.feat.keys():
            del self.lvl[i]
            del self.ptsi[i]
            del self.ptsm[i]
            del self.feat[i]

    def FineTuning(self, im=None):
        """Redefine and refine the points selected in the images.

        Parameters
        ----------
        im : int (OPTIONNAL)
            the feature number is only one feature has to be redefined. Default = all
            
        """

        # Arg: f pyxel image or Array of pyxel images
        if im is None:
            rg = self.ptsi.keys()
        else:
            rg = np.array([im])
        for i in rg:  # loop on features
            for j in range(len(self.ptsi[i][:, 1])):  # loop on points
                x = int(self.ptsi[i][j, 0])
                y = int(self.ptsi[i][j, 1])
                umin = max(0, x - 50)
                vmin = max(0, y - 50)
                umax = min(self.f.pix.shape[1] - 1, x + 50)
                vmax = min(self.f.pix.shape[0] - 1, y + 50)
                fsub = self.f.pix[vmin:vmax, umin:umax]
                plt.imshow(fsub, cmap="gray", interpolation="none")
                plt.plot(x - umin, y - vmin, "y+")
                figManager = plt.get_current_fig_manager()
                figManager.window.showMaximized()
                self.ptsi[i][j, :] = np.array(plt.ginput(1))[0] + np.array([umin, vmin])
                plt.close()

    def Calibration(self):
        """Performs the calibration provided that sufficient features have been 
        selected using NewPoint(), NewLine() or NewCircle().
            
        Returns
        -------
        pyxel Camera object
            The calibrated camera model    
        """
        # Compute Levelsets
        for i in self.feat.keys():
            if "circle" in self.feat[i]:
                self.lvl[i] = LSfromCircle(self.f, self.ptsi[i])
            elif "line" in self.feat[i]:
                self.lvl[i] = LSfromLine(self.f, self.ptsi[i])
            elif "point" in self.feat[i]:
                self.lvl[i] = LSfromPoint(self.f, self.ptsi[i])
        # Calibration
        xp = dict()
        yp = dict()
        for i in self.feat.keys():
            self.lvl[i].BuildInterp()
            xp[i] = self.m.n[self.ptsm[i], 0]
            yp[i] = self.m.n[self.ptsm[i], 1]
        if self.cam is None:
            if len(self.feat) > 2:
                ptsm = np.empty((0, 2))
                ptsM = np.empty((0, 2))
                for i in self.feat.keys():
                    ptsm = np.vstack((ptsm, np.mean(self.ptsi[i], axis=0)))
                    ptsM = np.vstack((ptsM, np.mean(self.m.n[self.ptsm[i]], axis=0)))
                self.Init3Pts(ptsm, ptsM)
            else:
                self.Init3Pts()
        p = self.cam.get_p()
        C = np.diag(p)
        if p[-1] == 0:
            C[-1, -1] = 1
        for i in range(40):
            M = np.zeros((len(p), len(p)))
            b = np.zeros(len(p))
            for j in self.feat.keys():
                up, vp = self.cam.P(xp[j], yp[j])
                lp = self.lvl[j].Interp(up, vp)
                dPudp, dPvdp = self.cam.dPdp(xp[j], yp[j])
                ldxr, ldyr = self.lvl[j].InterpGrad(up, vp)
                dPdl = np.diag(ldxr) @ dPudp + np.diag(ldyr).dot(dPvdp)
                M += C.T.dot(dPdl.T.dot(dPdl.dot(C)))
                b += C.T.dot(dPdl.T.dot(lp))
            dp = C.dot(np.linalg.solve(M, -b))
            p += 0.8 * dp
            self.cam.set_p(p)
            err = np.linalg.norm(dp) / np.linalg.norm(p)
            print("Iter # %2d | disc=%2.2f %% | dU/U=%1.2e"
                % (i + 1, np.mean(lp) / max(self.f.pix.shape) * 100, err))
            if err < 1e-5:
                break
        print("cam = px.Camera(np.array([%f, %f, %f, %f]))" % (p[0], p[1], p[2], p[3]))
        return self.cam




# -*- coding: utf-8 -*-
"""
Finite Element Digital Image Correlation method 

@author: JC Passieux, INSA Toulouse, 2021

pyxel

PYthon library for eXperimental mechanics using Finite ELements

"""

import numpy as np

class Camera:
    def __init__(self, p):
        self.set_p(p)

    def set_p(self, p):
        """
        Set manually the camera parameters.

        Parameters
        ----------
        p : NUMPY.ARRAY
            np.array([f, tx, ty, rz])
            f: focal length (scaling)
            tx: x-translation
            ty: y-translation
            rz: z-rotation

        """
        self.f = float(p[0])
        self.tx = float(p[1])
        self.ty = float(p[2])
        self.rz = float(p[3])

    def get_p(self):
        """
        Returns the vector of parameters of the camera model

        Returns
        -------
        NUMPY.ARRAY

        """
        return np.array([self.f, self.tx, self.ty, self.rz])

    def SubSampleCopy(self, nscale):
        """
        Camera model copy with subsampling (of bining) for multiscale initialization
        
        Parameters
        ----------
        NSCALE : INT
            number of scales, such that the scaling is 2**NSCALE

        Returns
        -------
        CAM : 1D PYXEL.CAMERA
            A new camera model that maps the same mesh coord. sys. to the  
            same image but with a bining of NSCALE. 
        """
        p = self.get_p()
        p[0] /= 2 ** nscale
        return Camera(p)

    def P(self, X, Y):
        """
        Camera model projection. Maps a point of the mesh to a point
        in the image plane

        Parameters
        ----------
        X : NUMPY.ARRAY
            X coordinate in the mesh system of the points to map
        Y : NUMPY.ARRAY
            Y coordinate in the mesh system of the points to map

        Returns
        -------
        u : 1D NUMPY.ARRAY
            u coordinate of the corresponding point in the image system
        v : 1D NUMPY.ARRAY
            v coordinate of the corresponding point in the image system

        """
        u = -self.f * (-np.sin(self.rz) * X + np.cos(self.rz) * Y + self.ty)
        v = self.f * (np.cos(self.rz) * X + np.sin(self.rz) * Y + self.tx)
        return u, v

    def Pinv(self, u, v):
        """
        Inverse of the Camera model. Maps a point in the image to a point
        in the mesh coordinate sys.
        (Explicit version for 2D camera model)

        Parameters
        ----------
        u : 1D NUMPY.ARRAY
            u coordinate in the image system of the points to map
        v : 1D NUMPY.ARRAY
            v coordinate in the image system of the points to map

        Returns
        -------
        X : 1D NUMPY.ARRAY
            X coordinate of the corresponding position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the corresponding position in the mesh system

        """
        X = -np.sin(self.rz) * (-u / self.f - self.ty) + np.cos(self.rz) * (
            v / self.f - self.tx
        )
        Y = np.cos(self.rz) * (-u / self.f - self.ty) + np.sin(self.rz) * (
            v / self.f - self.tx
        )
        return X, Y

    def PinvNL(self, u, v):
        """
        Inverse of the Camera model. Maps a point in the image to a point
        in the mesh coordinate sys.
        (General version for any possibly NL camera models)

        Parameters
        ----------
        u : 1D NUMPY.ARRAY
            u coordinate in the image system of the points to map
        v : 1D NUMPY.ARRAY
            v coordinate in the image system of the points to map

        Returns
        -------
        X : 1D NUMPY.ARRAY
            X coordinate of the corresponding position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the corresponding position in the mesh system

        """
        X = np.zeros(len(u))
        Y = np.zeros(len(u))
        for ii in range(10):
            pnx, pny = self.P(X, Y)
            resx = u - pnx
            resy = v - pny
            dPxdX, dPxdY, dPydX, dPydY = self.dPdX(X, Y)
            detJ = dPxdX * dPydY - dPxdY * dPydX
            dX = dPydY / detJ * resx - dPxdY / detJ * resy
            dY = -dPydX / detJ * resx + dPxdX / detJ * resy
            X += dX
            Y += dY
            res = np.linalg.norm(dX) + np.linalg.norm(dY)
            if res < 1e-4:
                break
        return X, Y

    def dPdX(self, X, Y):
        """
        Derivative of the Camera model wrt physical position X, Y

        Parameters
        ----------
        X : 1D NUMPY.ARRAY
            X coordinate of the current position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the current position in the mesh system

        Returns
        -------
        dudx : NUMPY.ARRAY
            Derivative of coord. u wrt to X
        dudy : NUMPY.ARRAY
            Derivative of coord. u wrt to Y
        dvdx : NUMPY.ARRAY
            Derivative of coord. v wrt to X
        dvdy : NUMPY.ARRAY
            Derivative of coord. v wrt to Y

        """
        dudx = self.f * np.sin(self.rz) * np.ones(X.shape[0])
        dudy = -self.f * np.cos(self.rz) * np.ones(X.shape[0])
        dvdx = self.f * np.cos(self.rz) * np.ones(X.shape[0])
        dvdy = self.f * np.sin(self.rz) * np.ones(X.shape[0])
        return dudx, dudy, dvdx, dvdy

    def dPdp(self, X, Y):
        """
        First order derivative of the Camera model wrt camera parameters p

        Parameters
        ----------
        X : 1D NUMPY.ARRAY
            X coordinate of the current position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the current position in the mesh system

        Returns
        -------
        dudp : NUMPY.ARRAY
            Derivative of coord. u wrt to p
        dvdp : NUMPY.ARRAY
            Derivative of coord. v wrt to p

        """
        dudf = -1 * (-np.sin(self.rz) * X + np.cos(self.rz) * Y + self.ty)
        dudtx = 0 * X
        dudty = 0 * X - self.f
        dudrz = self.f * (np.cos(self.rz) * X + np.sin(self.rz) * Y)
        dvdf = np.cos(self.rz) * X + np.sin(self.rz) * Y + self.tx
        dvdtx = 0 * X + self.f
        dvdty = 0 * X
        dvdrz = self.f * (-np.sin(self.rz) * X + np.cos(self.rz) * Y)
        return np.c_[dudf, dudtx, dudty, dudrz], np.c_[dvdf, dvdtx, dvdty, dvdrz]

    def d2Pdp2(self, X, Y):
        """
        Second order derivative of the Camera model wrt camera parameters p

        Parameters
        ----------
        X : 1D NUMPY.ARRAY
            X coordinate of the current position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the current position in the mesh system

        Returns
        -------
        d2udp2 : NUMPY.ARRAY
            second order derivative of coord. u wrt to p
        d2vdp2 : NUMPY.ARRAY
            second order derivative of coord. v wrt to p

        """
        d2udf2 = 0 * X
        d2udtx2 = 0 * X
        d2udty2 = 0 * X
        d2udrz2 = self.f * (-np.sin(self.rz) * X + np.cos(self.rz) * Y)
        d2udftx = 0 * X
        d2udfty = 0 * X - 1
        d2udfrz = np.cos(self.rz) * X + np.sin(self.rz) * Y
        d2udtxty = 0 * X
        d2udtxrz = 0 * X
        d2udtyrz = 0 * X
        d2vdf2 = 0 * X
        d2vdtx2 = 0 * X
        d2vdty2 = 0 * X
        d2vdrz2 = -self.f * (np.cos(self.rz) * X + np.sin(self.rz) * Y)
        d2vdftx = 0 * X + 1
        d2vdfty = 0 * X
        d2vdfrz = -np.sin(self.rz) * X + np.cos(self.rz) * Y
        d2vdtxty = 0 * X
        d2vdtxrz = 0 * X
        d2vdtyrz = 0 * X
        d2udp2 = np.c_[d2udf2, d2udtx2, d2udty2, d2udrz2, d2udftx, 
                       d2udfty, d2udfrz, d2udtxty, d2udtxrz, d2udtyrz]
        d2vdp2 = np.c_[d2vdf2, d2vdtx2, d2vdty2, d2vdrz2, d2vdftx,
                       d2vdfty, d2vdfrz, d2vdtxty, d2vdtxrz, d2vdtyrz]
        return d2udp2, d2vdp2

    def ImageFiles(self, fname, imnums):
        self.fname = fname
        self.imnums = imnums

class CameraNL:
    def __init__(self, p):
        self.set_p(p)

    def set_p(self, p):
        """
        Set manually the camera parameters.

        Parameters
        ----------
        p : NUMPY.ARRAY
            np.array([f, tx, ty, rz])
            f: focal length (scaling)
            tx: x-translation
            ty: y-translation
            rz: z-rotation
            u0: image center coord x, default Npix_x/2
            v0: image center coord y, default Npix_y/2
            r1: first order radial dist param, default 0
        """
        self.f = p[0]
        self.tx = p[1]
        self.ty = p[2]
        self.rz = p[3]
        self.u0 = p[4]
        self.v0 = p[5]
        self.r1 = p[6]
        self.r2 = p[7]
        self.r3 = p[8]

    def get_p(self):
        """
        Returns the vector of parameters of the camera model

        Returns
        -------
        NUMPY.ARRAY

        """
        return np.array([self.f, self.tx, self.ty, self.rz])

    def SubSampleCopy(self, nscale):
        """
        Camera model copy with subsampling (of bining) for multiscale initialization
        
        Parameters
        ----------
        NSCALE : INT
            number of scales, such that the scaling is 2**NSCALE

        Returns
        -------
        CAM : 1D PYXEL.CAMERA
            A new camera model that maps the same mesh coord. sys. to the  
            same image but with a bining of NSCALE. 
        """
        p = self.get_p()
        p[0] /= 2 ** nscale
        return CameraNL(p)

    def P(self, X, Y):
        """
        Camera model projection. Maps a point of the mesh to a point
        in the image plane

        Parameters
        ----------
        X : NUMPY.ARRAY
            X coordinate in the mesh system of the points to map
        Y : NUMPY.ARRAY
            Y coordinate in the mesh system of the points to map

        Returns
        -------
        u : 1D NUMPY.ARRAY
            u coordinate of the corresponding point in the image system
        v : 1D NUMPY.ARRAY
            v coordinate of the corresponding point in the image system

        """
        u = - self.f * (-np.sin(self.rz) * X + np.cos(self.rz) * Y + self.ty)
        v = self.f * (np.cos(self.rz) * X + np.sin(self.rz) * Y + self.tx)
       
        rho2 = (u-self.u0)**2 + (v-self.v0)**2
        ud = (u-self.u0) * (1 + self.r1*rho2 + self.r2*rho2**2 + self.r3*rho2**3) + self.u0
        vd = (v-self.v0) * (1 + self.r1*rho2 + self.r2*rho2**2 + self.r3*rho2**3) + self.v0

        return ud, vd

    def Pinv(self, u, v):
        """
        Inverse of the Camera model. Maps a point in the image to a point
        in the mesh coordinate sys.
        (General version for any possibly NL camera models)

        Parameters
        ----------
        u : 1D NUMPY.ARRAY
            u coordinate in the image system of the points to map
        v : 1D NUMPY.ARRAY
            v coordinate in the image system of the points to map

        Returns
        -------
        X : 1D NUMPY.ARRAY
            X coordinate of the corresponding position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the corresponding position in the mesh system

        """
        X = np.zeros(len(u))
        Y = np.zeros(len(u))
        for ii in range(10):
            pnx, pny = self.P(X, Y)
            resx = u - pnx
            resy = v - pny
            dPxdX, dPxdY, dPydX, dPydY = self.dPdX(X, Y)
            detJ = dPxdX * dPydY - dPxdY * dPydX
            dX = dPydY / detJ * resx - dPxdY / detJ * resy
            dY = -dPydX / detJ * resx + dPxdX / detJ * resy
            X += dX
            Y += dY
            res = np.linalg.norm(dX) + np.linalg.norm(dY)
            if res < 1e-4:
                break
        return X, Y

    def dPdX(self, X, Y):
        """
        Derivative of the Camera model wrt physical position X, Y

        Parameters
        ----------
        X : 1D NUMPY.ARRAY
            X coordinate of the current position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the current position in the mesh system

        Returns
        -------
        dudx : NUMPY.ARRAY
            Derivative of coord. u wrt to X
        dudy : NUMPY.ARRAY
            Derivative of coord. u wrt to Y
        dvdx : NUMPY.ARRAY
            Derivative of coord. v wrt to X
        dvdy : NUMPY.ARRAY
            Derivative of coord. v wrt to Y

        """
        # import sympy as sp
        # f, tx, ty, rz, u0, v0, r1, X, Y = sp.symbols('f, tx, ty, rz, u0, v0, r1, X, Y', real=True)
        # u = f * (sp.sin(rz) * X - sp.cos(rz) * Y - ty)
        # v = f * (sp.cos(rz) * X + sp.sin(rz) * Y + tx)
        # rho2 = (u-u0)**2 + (v-v0)**2
        # ud = (u - u0) * (1 + r1*rho2) + u0 
        # vd = (v - v0) * (1 + r1*rho2) + v0
        # sp.pycode(sp.diff(vd, Y))
        f = self.f
        tx = self.tx
        ty = self.ty
        rz = self.rz
        u0 = self.u0
        v0 = self.v0
        r1 = self.r1
        dudx = f*(r1*((f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)**2 + (f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)**2) + 1)*np.sin(rz) + r1*(f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)*(2*f*(f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)*np.sin(rz) + 2*f*(f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)*np.cos(rz))
        dudy = -f*(r1*((f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)**2 + (f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)**2) + 1)*np.cos(rz) + r1*(f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)*(-2*f*(f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)*np.cos(rz) + 2*f*(f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)*np.sin(rz))
        dvdx = f*(r1*((f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)**2 + (f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)**2) + 1)*np.cos(rz) + r1*(f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)*(2*f*(f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)*np.sin(rz) + 2*f*(f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)*np.cos(rz))
        dvdy = f*(r1*((f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)**2 + (f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)**2) + 1)*np.sin(rz) + r1*(f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)*(-2*f*(f*(X*np.sin(rz) - Y*np.cos(rz) - ty) - u0)*np.cos(rz) + 2*f*(f*(X*np.cos(rz) + Y*np.sin(rz) + tx) - v0)*np.sin(rz))
        return dudx, dudy, dvdx, dvdy

    def dPdp(self, X, Y):
        """
        First order derivative of the Camera model wrt camera parameters p

        Parameters
        ----------
        X : 1D NUMPY.ARRAY
            X coordinate of the current position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the current position in the mesh system

        Returns
        -------
        dudp : NUMPY.ARRAY
            Derivative of coord. u wrt to p
        dvdp : NUMPY.ARRAY
            Derivative of coord. v wrt to p

        """
        # import sympy as sp
        # f, tx, ty, rz, u0, v0, r1, X, Y = sp.symbols('f, tx, ty, rz, u0, v0, r1, X, Y', real=True)
        # u = sp.sin(rz) * X - sp.cos(rz) * Y - ty
        # v = sp.cos(rz) * X + sp.sin(rz) * Y + tx
        # rho2 = (u-u0)**2 + (v-v0)**2
        # ud = f * u * (1 + r1*rho2) 
        # vd = f * v * (1 + r1*rho2) 
        # sp.pycode(sp.diff(vd, r1))
        f = self.f
        tx = self.tx
        ty = self.ty
        rz = self.rz
        u0 = self.u0
        v0 = self.v0
        r1 = self.r1
        
        dudf = (r1*((X*np.sin(rz) - Y*np.cos(rz) - ty - u0)**2 + (X*np.cos(rz) + Y*np.sin(rz) + tx - v0)**2) + 1)*(X*np.sin(rz) - Y*np.cos(rz) - ty)
        dudtx = f*r1*(X*np.sin(rz) - Y*np.cos(rz) - ty)*(2*X*np.cos(rz) + 2*Y*np.sin(rz) + 2*tx - 2*v0)       
        dudty = f*r1*(X*np.sin(rz) - Y*np.cos(rz) - ty)*(-2*X*np.sin(rz) + 2*Y*np.cos(rz) + 2*ty + 2*u0) - f*(r1*((X*np.sin(rz) - Y*np.cos(rz) - ty - u0)**2 + (X*np.cos(rz) + Y*np.sin(rz) + tx - v0)**2) + 1)
        dudrz = f*r1*((-2*X*np.sin(rz) + 2*Y*np.cos(rz))*(X*np.cos(rz) + Y*np.sin(rz) + tx - v0) + (2*X*np.cos(rz) + 2*Y*np.sin(rz))*(X*np.sin(rz) - Y*np.cos(rz) - ty - u0))*(X*np.sin(rz) - Y*np.cos(rz) - ty) + f*(X*np.cos(rz) + Y*np.sin(rz))*(r1*((X*np.sin(rz) - Y*np.cos(rz) - ty - u0)**2 + (X*np.cos(rz) + Y*np.sin(rz) + tx - v0)**2) + 1)
        dudu0 = f*r1*(X*np.sin(rz) - Y*np.cos(rz) - ty)*(-2*X*np.sin(rz) + 2*Y*np.cos(rz) + 2*ty + 2*u0)
        dudv0 = f*r1*(X*np.sin(rz) - Y*np.cos(rz) - ty)*(-2*X*np.cos(rz) - 2*Y*np.sin(rz) - 2*tx + 2*v0)
        dudr1 = f*((X*np.sin(rz) - Y*np.cos(rz) - ty - u0)**2 + (X*np.cos(rz) + Y*np.sin(rz) + tx - v0)**2)*(X*np.sin(rz) - Y*np.cos(rz) - ty)
        
        dvdf = (r1*((X*np.sin(rz) - Y*np.cos(rz) - ty - u0)**2 + (X*np.cos(rz) + Y*np.sin(rz) + tx - v0)**2) + 1)*(X*np.cos(rz) + Y*np.sin(rz) + tx)
        dvdtx = f*r1*(X*np.cos(rz) + Y*np.sin(rz) + tx)*(2*X*np.cos(rz) + 2*Y*np.sin(rz) + 2*tx - 2*v0) + f*(r1*((X*np.sin(rz) - Y*np.cos(rz) - ty - u0)**2 + (X*np.cos(rz) + Y*np.sin(rz) + tx - v0)**2) + 1)
        dvdty = f*r1*(X*np.cos(rz) + Y*np.sin(rz) + tx)*(-2*X*np.sin(rz) + 2*Y*np.cos(rz) + 2*ty + 2*u0)
        dvdrz = f*r1*((-2*X*np.sin(rz) + 2*Y*np.cos(rz))*(X*np.cos(rz) + Y*np.sin(rz) + tx - v0) + (2*X*np.cos(rz) + 2*Y*np.sin(rz))*(X*np.sin(rz) - Y*np.cos(rz) - ty - u0))*(X*np.cos(rz) + Y*np.sin(rz) + tx) + f*(-X*np.sin(rz) + Y*np.cos(rz))*(r1*((X*np.sin(rz) - Y*np.cos(rz) - ty - u0)**2 + (X*np.cos(rz) + Y*np.sin(rz) + tx - v0)**2) + 1)
        dvdu0 = f*r1*(X*np.cos(rz) + Y*np.sin(rz) + tx)*(-2*X*np.sin(rz) + 2*Y*np.cos(rz) + 2*ty + 2*u0)
        dvdv0 = f*r1*(X*np.cos(rz) + Y*np.sin(rz) + tx)*(-2*X*np.cos(rz) - 2*Y*np.sin(rz) - 2*tx + 2*v0)
        dvdr1 = f*((X*np.sin(rz) - Y*np.cos(rz) - ty - u0)**2 + (X*np.cos(rz) + Y*np.sin(rz) + tx - v0)**2)*(X*np.cos(rz) + Y*np.sin(rz) + tx)
        return np.c_[dudf, dudtx, dudty, dudrz, dudu0, dudv0, dudr1], np.c_[dvdf, dvdtx, dvdty, dvdrz, dvdu0, dvdv0, dvdr1]

    def d2Pdp2(self, X, Y):
        """
        Second order derivative of the Camera model wrt camera parameters p

        Parameters
        ----------
        X : 1D NUMPY.ARRAY
            X coordinate of the current position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the current position in the mesh system

        Returns
        -------
        d2udp2 : NUMPY.ARRAY
            second order derivative of coord. u wrt to p
        d2vdp2 : NUMPY.ARRAY
            second order derivative of coord. v wrt to p

        """

        return None

    def ImageFiles(self, fname, imnums):
        self.fname = fname
        self.imnums = imnums

#%%
# import sympy as sp
# X, Y, Z, f, rx, ry, rz, tx, ty, tz = sp.symbols('X, Y, Z, f, rx, ry, rz, tx, ty, tz')
# T = np.array([[1, 0, 0, tx],
#               [0, 1, 0, ty],
#               [0, 0, 1, tz],
#               [0, 0, 0, 1]])
# R1 = np.array([[sp.cos(rz),-sp.sin(rz), 0, 0],
#                [sp.sin(rz), sp.cos(rz), 0, 0],
#                [0, 0, 1, 0],
#                [0, 0, 0, 1]])
# R2 = np.array([[1, 0, 0, 0],
#                [0, sp.cos(rx),-sp.sin(rx), 0],
#                [0, sp.sin(rx), sp.cos(rx), 0],
#                [0, 0, 0, 1]])
# R3 = np.array([[sp.cos(ry), 0 ,-sp.sin(ry), 0],
#                [0, 1, 0, 0],
#                [sp.sin(ry), 0, sp.cos(ry), 0],
#                [0, 0, 0, 1]])
# v = np.array([X, Y, Z, 1])
# uvw = f*T@R1@R2@R3@v
# u = uvw[0]
# v = uvw[1]
# w = uvw[2]

# print(sp.pycode(u))
# print(sp.pycode(v))
# print(sp.pycode(w))

# print(sp.pycode(u.diff(X, 1)))
# print(sp.pycode(u.diff(Y, 1)))
# print(sp.pycode(u.diff(Z, 1)))

# print(sp.pycode(v.diff(X, 1)))
# print(sp.pycode(v.diff(Y, 1)))
# print(sp.pycode(v.diff(Z, 1)))

# print(sp.pycode(w.diff(X, 1)))
# print(sp.pycode(w.diff(Y, 1)))
# print(sp.pycode(w.diff(Z, 1)))

# print(sp.pycode(u.diff(f, 1)))
# print(sp.pycode(u.diff(tx, 1)))
# print(sp.pycode(u.diff(ty, 1)))
# print(sp.pycode(u.diff(tz, 1)))
# print(sp.pycode(u.diff(rx, 1)))
# print(sp.pycode(u.diff(ry, 1)))
# print(sp.pycode(u.diff(rz, 1)))

# print(sp.pycode(v.diff(f, 1)))
# print(sp.pycode(v.diff(tx, 1)))
# print(sp.pycode(v.diff(ty, 1)))
# print(sp.pycode(v.diff(tz, 1)))
# print(sp.pycode(v.diff(rx, 1)))
# print(sp.pycode(v.diff(ry, 1)))
# print(sp.pycode(v.diff(rz, 1)))

# print(sp.pycode(w.diff(f, 1)))
# print(sp.pycode(w.diff(tx, 1)))
# print(sp.pycode(w.diff(ty, 1)))
# print(sp.pycode(w.diff(tz, 1)))
# print(sp.pycode(w.diff(rx, 1)))
# print(sp.pycode(w.diff(ry, 1)))
# print(sp.pycode(w.diff(rz, 1)))




class CameraVol:
    def __init__(self, p):
        self.set_p(p)

    def set_p(self, p):
        """
        Set manually the camera parameters.

        Parameters
        ----------
        p : NUMPY.ARRAY
            np.array([f, tx, ty, tz, rx, ry, rz])
            f: focal length (scaling)
            tx: x-translation
            ty: y-translation
            tz: z-translation
            rx: x-rotation
            ry: y-rotation
            rz: z-rotation

        """
        self.f = float(p[0])
        self.tx = float(p[1])
        self.ty = float(p[2])
        self.tz = float(p[3])
        self.rx = float(p[4])
        self.ry = float(p[5])
        self.rz = float(p[6])

    def get_p(self):
        """
        Returns the vector of parameters of the camera model

        Returns
        -------
        NUMPY.ARRAY

        """
        return np.array([self.f, self.tx, self.ty, self.tz, self.rx, self.ry, self.rz])

    def SubSampleCopy(self, nscale):
        """
        Camera model copy with subsampling (of bining) for multiscale initialization
        
        Parameters
        ----------
        NSCALE : INT
            number of scales, such that the scaling is 2**NSCALE

        Returns
        -------
        CAM : 1D PYXEL.CAMERA
            A new camera model that maps the same mesh coord. sys. to the  
            same image but with a bining of NSCALE. 
        """
        p = self.get_p()
        p[0] /= 2 ** nscale
        return CameraVol(p)

    def P(self, X, Y, Z):
        """
        Camera model projection. Maps a point of the mesh to a point
        in the image plane

        Parameters
        ----------
        X : NUMPY.ARRAY
            X coordinate in the mesh system of the points to map
        Y : NUMPY.ARRAY
            Y coordinate in the mesh system of the points to map
        Z : NUMPY.ARRAY
            Z coordinate in the mesh system of the points to map

        Returns
        -------
        u : 1D NUMPY.ARRAY
            u coordinate of the corresponding point in the image system
        v : 1D NUMPY.ARRAY
            v coordinate of the corresponding point in the image system
        w : 1D NUMPY.ARRAY
            w coordinate of the corresponding point in the image system

        """
        f = self.f
        tx = self.tx
        ty = self.ty
        tz = self.tz
        rx = self.rx
        ry = self.ry
        rz = self.rz
        u = X*(f*np.sin(rx)*np.sin(ry)*np.sin(rz) + f*np.cos(ry)*np.cos(rz)) - \
            Y*f*np.sin(rz)*np.cos(rx) + Z*(f*np.sin(rx)*np.sin(rz)*np.cos(ry) - \
                                                 f*np.sin(ry)*np.cos(rz)) + f*tx
        v = X*(-f*np.sin(rx)*np.sin(ry)*np.cos(rz) + f*np.sin(rz)*np.cos(ry)) + \
            Y*f*np.cos(rx)*np.cos(rz) + Z*(-f*np.sin(rx)*np.cos(ry)*np.cos(rz) - \
                                               f*np.sin(ry)*np.sin(rz)) + f*ty
        w = X*f*np.sin(ry)*np.cos(rx) + Y*f*np.sin(rx) + \
            Z*f*np.cos(rx)*np.cos(ry) + f*tz
        return u, v, w

    # import sympy as sp
    # dPxdX, dPxdY, dPxdZ, dPydX, dPydY, dPydZ, dPzdX, dPzdY, dPzdZ = \
    #    sp.symbols('dPxdX, dPxdY, dPxdZ, dPydX, dPydY, dPydZ, dPzdX, dPzdY, dPzdZ')    
    # J = sp.Matrix([[dPxdX, dPxdY, dPxdZ],
    #                [dPydX, dPydY, dPydZ],
    #                [dPzdX, dPzdY, dPzdZ]])
    # detJ = sp.det(J)
    # print(sp.pycode(detJ))
    # Jinv = J.inv()*detJ
    # print(Jinv[0,0])
    # print(Jinv[0,1])
    # print(Jinv[0,2])
    # print(Jinv[1,0])
    # print(Jinv[1,1])
    # print(Jinv[1,2])    
    # print(Jinv[2,0])
    # print(Jinv[2,1])
    # print(Jinv[2,2])

    def Pinv(self, u, v, w):
        """
        Inverse of the Camera model. Maps a point in the image to a point
        in the mesh coordinate sys.
        (General version for any possibly NL camera models)

        Parameters
        ----------
        u : 1D NUMPY.ARRAY
            u coordinate in the image system of the points to map
        v : 1D NUMPY.ARRAY
            v coordinate in the image system of the points to map
        w : 1D NUMPY.ARRAY
            w coordinate in the image system of the points to map

        Returns
        -------
        X : 1D NUMPY.ARRAY
            X coordinate of the corresponding position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the corresponding position in the mesh system
        Z : 1D NUMPY.ARRAY
            Z coordinate of the corresponding position in the mesh system

        """
        X = np.zeros_like(u)
        Y = np.zeros_like(u)
        Z = np.zeros_like(u)
        for ii in range(10):
            pnx, pny, pnz = self.P(X, Y, Z)
            resx = u - pnx
            resy = v - pny
            resz = w - pnz
            dPxdX, dPxdY, dPxdZ, dPydX, dPydY, dPydZ, dPzdX, dPzdY, dPzdZ = self.dPdX(X, Y, Z)
            detJ = dPxdX*dPydY*dPzdZ - dPxdX*dPydZ*dPzdY - dPxdY*dPydX*dPzdZ + \
                dPxdY*dPydZ*dPzdX + dPxdZ*dPydX*dPzdY - dPxdZ*dPydY*dPzdX
            dX = ((dPydY*dPzdZ - dPydZ*dPzdY) * resx + \
                (-dPxdY*dPzdZ + dPxdZ*dPzdY) * resy + \
                 (dPxdY*dPydZ - dPxdZ*dPydY) * resz) / detJ
            dY = ((-dPydX*dPzdZ + dPydZ*dPzdX) * resx + \
                   (dPxdX*dPzdZ - dPxdZ*dPzdX) * resy + \
                  (-dPxdX*dPydZ + dPxdZ*dPydX) * resz ) / detJ
            dZ = ((dPydX*dPzdY - dPydY*dPzdX) * resx + \
                 (-dPxdX*dPzdY + dPxdY*dPzdX) * resy + \
                  (dPxdX*dPydY - dPxdY*dPydX) * resz ) / detJ
            X += dX
            Y += dY
            Z += dZ
            res = np.linalg.norm(dX) + np.linalg.norm(dY) + np.linalg.norm(dZ)
            if res < 1e-4:
                break
        return X, Y, Z

    def dPdX(self, X, Y, Z):
        """
        Derivative of the Camera model wrt physical position X, Y

        Parameters
        ----------
        X : 1D NUMPY.ARRAY
            X coordinate of the current position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the current position in the mesh system
        Z : 1D NUMPY.ARRAY
            Z coordinate of the current position in the mesh system

        Returns
        -------
        dudx : NUMPY.ARRAY
            Derivative of coord. u wrt to X
        dudy : NUMPY.ARRAY
            Derivative of coord. u wrt to Y
        dudz : NUMPY.ARRAY
            Derivative of coord. u wrt to Z
        dvdx : NUMPY.ARRAY
            Derivative of coord. v wrt to X
        dvdy : NUMPY.ARRAY
            Derivative of coord. v wrt to Y
        dvdz : NUMPY.ARRAY
            Derivative of coord. v wrt to Z
        dwdx : NUMPY.ARRAY
            Derivative of coord. w wrt to X
        dwdy : NUMPY.ARRAY
            Derivative of coord. w wrt to Y
        dwdz : NUMPY.ARRAY
            Derivative of coord. w wrt to Z

        """
        f = self.f
        rx = self.rx
        ry = self.ry
        rz = self.rz
        ones = np.ones(X.shape[0])
        dudx = (f*np.sin(rx)*np.sin(ry)*np.sin(rz) + f*np.cos(ry)*np.cos(rz)) * ones
        dudy = -f*np.sin(rz)*np.cos(rx)*ones
        dudz = (f*np.sin(rx)*np.sin(rz)*np.cos(ry) - f*np.sin(ry)*np.cos(rz))*ones
        dvdx = (-f*np.sin(rx)*np.sin(ry)*np.cos(rz) + f*np.sin(rz)*np.cos(ry))*ones
        dvdy = f*np.cos(rx)*np.cos(rz)*ones
        dvdz = (-f*np.sin(rx)*np.cos(ry)*np.cos(rz) - f*np.sin(ry)*np.sin(rz))*ones
        dwdx = f*np.sin(ry)*np.cos(rx)*ones
        dwdy = f*np.sin(rx)*ones
        dwdz = f*np.cos(rx)*np.cos(ry)*ones
        return dudx, dudy, dudz, dvdx, dvdy, dvdz, dwdx, dwdy, dwdz

    def dPdp(self, X, Y, Z):
        """
        First order derivative of the Camera model wrt camera parameters p

        Parameters
        ----------
        X : 1D NUMPY.ARRAY
            X coordinate of the current position in the mesh system
        Y : 1D NUMPY.ARRAY
            Y coordinate of the current position in the mesh system
        Z : 1D NUMPY.ARRAY
            Z coordinate of the current position in the mesh system

        Returns
        -------
        dudp : NUMPY.ARRAY
            Derivative of coord. u wrt to p
        dvdp : NUMPY.ARRAY
            Derivative of coord. v wrt to p
        dwdp : NUMPY.ARRAY
            Derivative of coord. w wrt to p

        """
        f = self.f
        tx = self.tx
        ty = self.ty
        tz = self.tz
        rx = self.rx
        ry = self.ry
        rz = self.rz
        #
        dudf = X*(np.sin(rx)*np.sin(ry)*np.sin(rz) + np.cos(ry)*np.cos(rz)) - \
            Y*np.sin(rz)*np.cos(rx) + Z*(np.sin(rx)*np.sin(rz)*np.cos(ry) - \
                                             np.sin(ry)*np.cos(rz)) + tx
        dudtx = f + 0 * X
        dudty = 0 * X
        dudtz = 0 * X
        dudrx = X*f*np.sin(ry)*np.sin(rz)*np.cos(rx) + Y*f*np.sin(rx)*np.sin(rz) + \
            Z*f*np.sin(rz)*np.cos(rx)*np.cos(ry)
        dudry = X*(f*np.sin(rx)*np.sin(rz)*np.cos(ry) - f*np.sin(ry)*np.cos(rz)) + \
            Z*(-f*np.sin(rx)*np.sin(ry)*np.sin(rz) - f*np.cos(ry)*np.cos(rz))
        dudrz = X*(f*np.sin(rx)*np.sin(ry)*np.cos(rz) - f*np.sin(rz)*np.cos(ry)) - \
            Y*f*np.cos(rx)*np.cos(rz) + Z*(f*np.sin(rx)*np.cos(ry)*np.cos(rz) + \
                                               f*np.sin(ry)*np.sin(rz))
        dvdf = X*(-np.sin(rx)*np.sin(ry)*np.cos(rz) + np.sin(rz)*np.cos(ry)) + \
            Y*np.cos(rx)*np.cos(rz) + Z*(-np.sin(rx)*np.cos(ry)*np.cos(rz) - \
                                             np.sin(ry)*np.sin(rz)) + ty
        # 
        dvdtx = 0 * X
        dvdty = 0 * X + f
        dvdtz = 0 * X
        dvdrx = -X*f*np.sin(ry)*np.cos(rx)*np.cos(rz) - Y*f*np.sin(rx)*np.cos(rz) - \
            Z*f*np.cos(rx)*np.cos(ry)*np.cos(rz)
        dvdry = X*(-f*np.sin(rx)*np.cos(ry)*np.cos(rz) - f*np.sin(ry)*np.sin(rz)) + \
            Z*(f*np.sin(rx)*np.sin(ry)*np.cos(rz) - f*np.sin(rz)*np.cos(ry))
        dvdrz = X*(f*np.sin(rx)*np.sin(ry)*np.sin(rz) + f*np.cos(ry)*np.cos(rz)) - \
            Y*f*np.sin(rz)*np.cos(rx) + Z*(f*np.sin(rx)*np.sin(rz)*np.cos(ry) - \
                                               f*np.sin(ry)*np.cos(rz))
        #                 
        dwdf = X*np.sin(ry)*np.cos(rx) + Y*np.sin(rx) + Z*np.cos(rx)*np.cos(ry) + tz
        dwdtx = 0 * X
        dwdty = 0 * X
        dwdtz = 0 * X + f
        dwdrx = -X*f*np.sin(rx)*np.sin(ry) + Y*f*np.cos(rx) - Z*f*np.sin(rx)*np.cos(ry)
        dwdry = X*f*np.cos(rx)*np.cos(ry) - Z*f*np.sin(ry)*np.cos(rx)
        dwdrz = 0 * X 
        #
        dudp = np.c_[dudf, dudtx, dudty, dudtz, dudrx, dudry, dudrz]
        dvdp = np.c_[dvdf, dvdtx, dvdty, dvdtz, dvdrx, dvdry, dvdrz]
        dwdp = np.c_[dwdf, dwdtx, dwdty, dwdtz, dwdrx, dwdry, dwdrz]        
        return dudp, dvdp, dwdp

    def ImageFiles(self, fname, imnums):
        self.fname = fname
        self.imnums = imnums









""" ************       USELESS 23/3/2022 !!!    *************** """









import numpy as np
from .mesh import Mesh


#%%
def ReadMeshGMSH(fn, dim=2):
    """
    Beta GMSH parser and converter to pyxel.Mesh object.
    
    Parameters
    ----------
        fn : string (filename)
        dim : int (DEFAULT dim=2)
    """
    mshfid = open(fn, "r")
    line = mshfid.readline()
    while line.find("$Nodes") < 0:
        line = mshfid.readline()
        pass
    line = mshfid.readline()
    nnodes = int(line)
    nodes = np.zeros((nnodes, 3))
    for jn in range(nnodes):
        sl = mshfid.readline().split()
        nodes[jn] = np.double(sl[1:])
    while line.find("$Elements") < 0:
        line = mshfid.readline()
        pass
    line = mshfid.readline()
    nelems = int(line)
    elems = dict()
    ne = 1
    line = np.int32(mshfid.readline().split())
    while ne < nelems:
        et = line[1]
        if et == 3:  # qua4
            nn = 4
            rep = np.arange(5, 9)
        elif et == 2:  # tri3
            nn = 3
            rep = np.arange(5, 8)
        elif et == 15:  # point
            nn = 1
            rep = [4]
        elif et == 1:  # segment
            nn = 2
            rep = np.arange(5, 7)
        elif et == 8:  # quadratic segment
            nn = 3
            rep = np.arange(5, 8)
        elif et == 9:  # tri6
            nn = 6
            rep = np.arange(5, 11)
        elif et == 16:  # qua8
            nn = 8
            rep = np.arange(5, 13)
        elif et == 10:  # qua9
            nn = 9
            rep = np.arange(5, 14)
        elems[et] = np.empty((0, nn), dtype=int)
        while line[1] == et:
            elems[et] = np.vstack((elems[et], line[rep] - 1))
            try:
                line = np.int32(mshfid.readline().split())
            except:
                break
            ne += 1
    if dim == 2:
        nodes = np.delete(nodes, 2, 1)
    if 15 in elems.keys():
        del elems[15]  # remove points
    if 1 in elems.keys():
        del elems[1]  # remove segments
    if 8 in elems.keys():
        del elems[8]  # remove quad segments
    m = Mesh(elems, nodes, dim)
    return m

def ReadMeshINP(fn):
    """
    Beta ABAQUS INP parser with 2D mesh and converter to pyxel.Mesh object.
    
    Parameters
    ----------
        fn: string (filename)
    
    2D ONLY : any type of S4, E4, S3, E3, S6, E6, S8, E8
    """

    mshfid = open(fn, "r")
    line = mshfid.readline()
    while line.find("*Node") < 0:
        line = mshfid.readline()
        pass
    line = mshfid.readline()
    dim = len(line.split(","))-1
    nodes = np.zeros((0, dim))
    while line.find("*Element") < 0:
        nodes = np.vstack((nodes, np.double(line.split(",")[1:])))
        line = mshfid.readline()
    elems = dict()
    while "*Element" in line:  # Loop on different element types
        print(line[:-1])
        if "S4" in line or "E4" in line:
            et = 3
            nn = 4
            rep = [1, 2, 3, 4]
        elif "S3" in line or "E3" in line:
            et = 2
            nn = 3
            rep = [1, 2, 3]
        elif "S6" in line or "E6" in line:
            et = 9
            nn = 6
            rep = [1, 2, 3, 4, 5, 6]
        elif "S8" in line or "E8" in line:
            et = 16
            nn = 8
            rep = [1, 2, 3, 4, 5, 6, 7, 8]
        elems[et] = np.empty((0, nn), dtype=int)
        while True:
            line = mshfid.readline().split(",")
            try:
                line = np.int32(line)
            except:
                break
            elems[et] = np.vstack((elems[et], line[rep] - 1))
    m = Mesh(elems, nodes)
    return m

def ReadMeshINPwithElset(fn):
    """
    Beta ABAQUS INP parser with 2D mesh and converter to pyxel.Mesh object.
    Exports also the element sets.

    Parameters
    ----------
        fn : string (filename)
    
    2D ONLY : any type of S4, E4, S3, E3, S6, E6, S8, E8
    """
    mshfid = open(fn, "r")
    line = mshfid.readline()
    while line.find("*Node") < 0:
        line = mshfid.readline()
        pass
    nodes = np.zeros((0, 2))
    line = mshfid.readline()
    while line.find("*Element") < 0:
        nodes = np.vstack((nodes, np.double(line.split(",")[1:])))
        line = mshfid.readline()
    # nnodes = nodes.shape[0]
    elems = dict()
    while "*Element" in line:  # Loop on different element types
        print(line[:-1])
        if "S4" in line or "E4" in line:
            et = 3
            nn = 4
            rep = [1, 2, 3, 4]
        elif "S3" in line or "E3" in line:
            et = 2
            nn = 3
            rep = [1, 2, 3]
        elif "S6" in line or "E6" in line:
            et = 9
            nn = 6
            rep = [1, 2, 3, 4, 5, 6]
        elif "S8" in line or "E8" in line:
            et = 16
            nn = 8
            rep = [1, 2, 3, 4, 5, 6, 7, 8]
        else:
            print("Unknown Element!")
            print(line)
        elems[et] = np.empty((0, nn), dtype=int)
        while True:
            line = mshfid.readline()
            try:
                line = np.int32(line.split(","))
            except:
                break
            elems[et] = np.vstack((elems[et], line[rep] - 1))
    elset = dict()
    nelset = 0
    while line.find("*") >= 0:
        if line.find("*Elset") >= 0:
            print(line[:-1])
            if line.find("generate") >= 0:
                line = mshfid.readline()
                gen = np.int32(line.split(","))
                elset[nelset] = np.arange(gen[0] - 1, gen[1], gen[2])
                line = mshfid.readline()
            else:
                line = mshfid.readline()
                lineconcat = ""
                while line.find("*") < 0:
                    lineconcat += "," + line
                    line = mshfid.readline()
                if lineconcat[-2] == ",":
                    lineconcat = lineconcat[:-2]
                elset[nelset] = np.int32(lineconcat[1:].split(",")) - 1
            nelset += 1
        elif line.find("*End Part") >= 0:
            break
        else:
            line = mshfid.readline()
            while line.find("*") < 0:
                line = mshfid.readline()
    m = Mesh(elems, nodes)
    return m, elset


def ReadMeshINP3D(fn):
    """BETA"""
    lines = open(fn, "r").readlines()
    k = 0
    while lines[k] != "*Node\r\n":
        k += 1
    k += 1
    nodes = np.zeros((0, 3))
    while lines[k][0:8] != "*Element":
        nodes = np.vstack((nodes, np.fromstring(lines[k], sep=",")[1:]))
        k += 1
    # here lines[k] == '*Element, type=C3D8R\r\n'
    k += 1
    elems = np.zeros((0, 9), dtype="int")
    while lines[k][0:1] != "*":
        elems = np.vstack((elems, np.fromstring(lines[k], sep=",", dtype="int") - 1))
        k += 1
    elems[:, 0] = 5
    m = Mesh(elems, nodes)
    return m





# -*- coding: utf-8 -*-
"""
Created on Sat Jan 29 08:16:13 2022

@author: passieux
"""
import numpy as np
import matplotlib.pyplot as plt
import gmsh
from .mesh import ReadMesh
from .camera import Camera
from .image import Image
from .mesh import Mesh
from skimage import measure

#%%


def StructuredMeshQ4(box, dx):
    """Build a structured linear Q4 mesh from two points coordinates (box)
    box = np.array([[xmin, ymin],
                    [xmax, ymax]])   in mesh unit
    dx = [dx, dy]: average element size (can be scalar)  in mesh unit"""
    dbox = box[1] - box[0]
    NE = (dbox / dx).astype(np.int64)
    NE = np.max(np.c_[NE, np.ones(2, dtype=int)], axis=1)
    X, Y = np.meshgrid(np.linspace(box[0, 0], box[1, 0], NE[0] + 1),
                       np.linspace(box[0, 1], box[1, 1], NE[1] + 1))
    n = np.vstack((X.T.ravel(), Y.T.ravel())).T
    n1 = np.kron(np.ones(NE[0]), np.arange(NE[1])) + \
        np.kron(np.arange(NE[0]), np.ones(NE[1])*(NE[1]+1))
    n2 = np.kron(np.ones(NE[0]), np.arange(NE[1])) + \
        np.kron(np.arange(1, NE[0]+1), np.ones(NE[1])*(NE[1]+1))
    n3 = np.kron(np.ones(NE[0]), np.arange(1, NE[1]+1)) + \
        np.kron(np.arange(1, NE[0]+1), np.ones(NE[1])*(NE[1]+1))
    n4 = np.kron(np.ones(NE[0]), np.arange(1, NE[1]+1)) + \
        np.kron(np.arange(NE[0]), np.ones(NE[1])*(NE[1]+1))
    e = np.c_[n1, n2, n3, n4].astype(int)
    m = Mesh({3: e}, n)
    return m


def StructuredMeshQ9(box, dx):
    """Build a structured quadratic Q9 mesh from two points coordinates (box)
    box = np.array([[xmin, ymin],
                    [xmax, ymax]])    in mesh unit
    dx = [dx, dy]: average element size (can be scalar)  in mesh unit"""
    dbox = box[1] - box[0]
    NE = (dbox / dx).astype(np.int64)
    NE = np.max(np.c_[NE, np.ones(2, dtype=int)], axis=1)
    X, Y = np.meshgrid(np.linspace(box[0, 0], box[1, 0], 2 * NE[0] + 1),
                       np.linspace(box[0, 1], box[1, 1], 2 * NE[1] + 1))
    n = np.vstack((X.T.ravel(), Y.T.ravel())).T
    nel = np.prod(NE)
    e = np.zeros((nel, 9), dtype=int)
    for ix in range(NE[0]):
        for iy in range(NE[1]):
            p1 = 2 * ix * (2 * NE[1] + 1) + 2 * iy
            p8 = p1 + 1
            p4 = p1 + 2
            p5 = 2 * ix * (2 * NE[1] + 1) + 2 * iy + 2 * NE[1] + 1
            p9 = p5 + 1
            p7 = p5 + 2
            p2 = 2 * ix * (2 * NE[1] + 1) + 2 * iy + 2 * (2 * NE[1] + 1)
            p6 = p2 + 1
            p3 = p2 + 2
            e[ix * NE[1] + iy,
                :] = np.array([p1, p2, p3, p4, p5, p6, p7, p8, p9])
    el = {10: e}
    m = Mesh(el, n)
    return m


def StructuredMeshQ8(box, dx):
    """Build a structured quadratic Q8 mesh from two points coordinates (box)
    box = np.array([[xmin, ymin],
                    [xmax, ymax]])   in mesh unit
    dx = [dx, dy]: average element size (can be scalar)  in mesh unit"""
    dbox = box[1] - box[0]
    NE = (dbox / dx).astype(np.int64)
    NE = np.max(np.c_[NE, np.ones(2, dtype=int)], axis=1)
    X, Y = np.meshgrid(np.linspace(box[0, 0], box[1, 0], 2 * NE[0] + 1),
                       np.linspace(box[0, 1], box[1, 1], 2 * NE[1] + 1))
    n = np.vstack((X.T.ravel(), Y.T.ravel())).T
    nel = np.prod(NE)
    e = np.zeros((nel, 8), dtype=int)
    for ix in range(NE[0]):
        for iy in range(NE[1]):
            p1 = 2 * ix * (2 * NE[1] + 1) + 2 * iy
            p8 = p1 + 1
            p4 = p1 + 2
            p5 = 2 * ix * (2 * NE[1] + 1) + 2 * iy + 2 * NE[1] + 1
            p7 = p5 + 2
            p2 = 2 * ix * (2 * NE[1] + 1) + 2 * iy + 2 * (2 * NE[1] + 1)
            p6 = p2 + 1
            p3 = p2 + 2
            n[p5 + 1, :] = n[0, :] * np.nan
            e[ix * NE[1] + iy] = np.array([p1, p2, p3, p4, p5, p6, p7, p8])
    el = {16: e}
    m = Mesh(el, n)
    return m


def StructuredMeshT3(box, dx):
    """Build a structured linear T3 mesh from two points coordinates (box)
    box = np.array([[xmin, ymin],
                    [xmax, ymax]])    in mesh unit
    dx = [dx, dy]: average element size (can be scalar) in mesh unit"""
    dbox = box[1] - box[0]
    NE = (dbox / dx).astype(np.int64)
    NE = np.max(np.c_[NE, np.ones(2, dtype=int)], axis=1)
    X, Y = np.meshgrid(np.linspace(box[0, 0], box[1, 0], NE[0] + 1),
                       np.linspace(box[0, 1], box[1, 1], NE[1] + 1))
    n = np.vstack((X.T.ravel(), Y.T.ravel())).T
    n1 = np.kron(np.ones(NE[0]), np.arange(NE[1])) + \
        np.kron(np.arange(NE[0]), np.ones(NE[1])*(NE[1]+1))
    n2 = np.kron(np.ones(NE[0]), np.arange(NE[1])) + \
        np.kron(np.arange(1, NE[0]+1), np.ones(NE[1])*(NE[1]+1))
    n3 = np.kron(np.ones(NE[0]), np.arange(1, NE[1]+1)) + \
        np.kron(np.arange(1, NE[0]+1), np.ones(NE[1])*(NE[1]+1))
    n4 = np.kron(np.ones(NE[0]), np.arange(1, NE[1]+1)) + \
        np.kron(np.arange(NE[0]), np.ones(NE[1])*(NE[1]+1))
    e = np.vstack((np.c_[n1, n2, n4], np.c_[n2, n3, n4])).astype(int)
    m = Mesh({2: e}, n)
    return m


def StructuredMeshT6(box, dx):
    """Build a structured quadratic T6 mesh from two points coordinates (box)
    box = np.array([[xmin, ymin],
                    [xmax, ymax]])    in mesh unit
    dx = [dx, dy]: average element size (can be scalar) in mesh unit"""
    dbox = box[1] - box[0]
    NE = (dbox / dx).astype(np.int64)
    X, Y = np.meshgrid(np.linspace(box[0, 0], box[1, 0], 2 * NE[0] + 1),
                       np.linspace(box[0, 1], box[1, 1], 2 * NE[1] + 1))
    n = np.vstack((X.T.ravel(), Y.T.ravel())).T
    nel = np.prod(NE) * 2
    e = np.zeros((nel, 6), dtype=int)
    for ix in range(NE[0]):
        for iy in range(NE[1]):
            p1 = 2 * ix * (2 * NE[1] + 1) + 2 * iy
            p8 = p1 + 1
            p4 = p1 + 2
            p5 = 2 * ix * (2 * NE[1] + 1) + 2 * iy + 2 * NE[1] + 1
            p9 = p5 + 1
            p7 = p5 + 2
            p2 = 2 * ix * (2 * NE[1] + 1) + 2 * iy + 2 * (2 * NE[1] + 1)
            p6 = p2 + 1
            p3 = p2 + 2
            e[2 * (ix * NE[1] + iy), :] = np.array([p1, p2, p4, p5, p9, p8])
            e[2 * (ix * NE[1] + iy) + 1, :] = np.array([p2, p3, p4, p6, p7, p9])
    el = {9: e}
    m = Mesh(el, n)
    return m


def StructuredMeshHex8(box, lc):
    """
    Meshes a box with hexahedral elements

    Parameters
    ----------
    box = np.array([[xmin, ymin, zmin],
                    [xmax, ymax, zmax]])    in mesh unit
    lc : mesh length.

    Returns
    -------
    m : Structured C8 mesh.

    """
    if type(lc) is int:
        lc = [lc, lc, lc]
    # Create a structured hexahedral mesh
    lx = box[1, 0] - box[0, 0]
    ly = box[1, 1] - box[0, 1]
    lz = box[1, 2] - box[0, 2]

    Nx = int(np.round(lx / lc[0]))
    Ny = int(np.round(ly / lc[1]))
    Nz = int(np.round(lz / lc[2]))

    x = np.linspace(0, lx, Nx+1)
    y = np.linspace(0, ly, Ny+1)
    z = np.linspace(0, lz, Nz+1)

    nbf_xi = len(x)
    nbf_eta = len(y)
    nbf_zeta = len(z)

    Xtot = np.kron(np.ones(nbf_eta*nbf_zeta), x)
    Ytot = np.kron(np.ones(nbf_zeta), np.kron(y, np.ones(nbf_xi)))
    Ztot = np.kron(z, np.ones(nbf_eta*nbf_xi))

    nodes = np.c_[Xtot, Ytot, Ztot]

    # Creates element connectivity
    p = 1
    q = 1
    r = 1
    nxi, neta, nzeta = Nx, Ny, Nz
    bf_xi_index = (np.kron(np.ones(nxi), np.arange(p+1)) +
                   np.kron(np.arange(nxi), np.ones(p+1))).reshape((nxi, p+1))
    bf_eta_index = (np.kron(np.ones(neta), np.arange(q+1)) +
                    np.kron(np.arange(neta), np.ones(q+1))).reshape((neta, q+1))
    bf_zeta_index = (np.kron(np.ones(nzeta), np.arange(r+1)) +
                     np.kron(np.arange(nzeta), np.ones(r+1))).reshape((nzeta, r+1))

    noelem = np.kron(np.ones_like(bf_zeta_index),  np.kron(np.ones_like(bf_eta_index), bf_xi_index)) + \
        (nxi+p)*np.kron(np.ones_like(bf_zeta_index), np.kron(bf_eta_index, np.ones_like(bf_xi_index))) + \
        (nxi+p)*(neta+q)*np.kron(bf_zeta_index,
                                 np.kron(np.ones_like(bf_eta_index), np.ones_like(bf_xi_index)))

    els = noelem.astype('int32')
    # In order to get the correct ordering of Meshio
    els = els[:, [2, 0, 1, 3, 6, 4, 5, 7]]
    # els = els[:,[]] # In order to get tge correct ordering of Gmsh
    e = {5: els}
    m = Mesh(e, nodes, 3)

    m.n[:, 0] += box[0, 0]
    m.n[:, 1] += box[0, 1]
    m.n[:, 2] += box[0, 2]

    return m


def StructuredMeshHex20(box, lc):
    """
    Meshes a box with hexahedral elements

    Parameters
    ----------
    box = np.array([[xmin, ymin, zmin],
                    [xmax, ymax, zmax]])    in mesh unit
    lc : mesh length.

    Returns
    -------
    m : Structured C20 mesh.

    """
    if type(lc) is int:
        lc = [lc, lc, lc]
    # Create a structured hexahedral mesh
    x0 = box[0, 0]
    y0 = box[0, 1]
    z0 = box[0, 2]
    lx = box[1, 0] - x0
    ly = box[1, 1] - y0
    lz = box[1, 2] - z0
    nx = int(lx/lc[0])
    ny = int(ly/lc[1])
    nz = int(lz/lc[2])
    gmsh.initialize()
    gmsh.model.add("lug")
    gmsh.model.geo.addPoint(x0, y0, z0, tag=1)
    gmsh.model.geo.addPoint(x0+lx, y0, z0, tag=2)
    gmsh.model.geo.addPoint(x0+lx, y0+ly, z0, tag=3)
    gmsh.model.geo.addPoint(x0, y0+ly, z0, tag=4)
    gmsh.model.geo.addLine(1, 2, tag=5)
    gmsh.model.geo.mesh.setTransfiniteCurve(5, nx+1)
    gmsh.model.geo.addLine(2, 3, tag=6)
    gmsh.model.geo.mesh.setTransfiniteCurve(6, ny+1)
    gmsh.model.geo.addLine(3, 4, tag=7)
    gmsh.model.geo.mesh.setTransfiniteCurve(7, nx+1)
    gmsh.model.geo.addLine(4, 1, tag=8)
    gmsh.model.geo.mesh.setTransfiniteCurve(8, ny+1)
    gmsh.model.geo.addCurveLoop([5, 6, 7, 8], 1)
    gmsh.model.geo.addPlaneSurface([1], 1)
    gmsh.model.geo.mesh.setTransfiniteSurface(1, "Left", [1, 2, 3, 4])
    gmsh.option.setNumber('Mesh.RecombineAll', 1)
    gmsh.option.setNumber('Mesh.RecombinationAlgorithm', 1)
    gmsh.option.setNumber('Mesh.Recombine3DLevel', 2)
    gmsh.option.setNumber('Mesh.SecondOrderIncomplete', 1)
    gmsh.model.geo.extrude([(2, 1)], 0, 0, lz, [nz, ], recombine=True)
    gmsh.option.setNumber('Mesh.ElementOrder', 2)
    gmsh.option.setNumber('General.Verbosity', 1)
    gmsh.model.geo.synchronize()
    gmsh.model.mesh.generate(3)
    gmsh.write("lug.msh")
    # if '-nopopup' not in sys.argv:
    #     gmsh.fltk.run()
    gmsh.finalize()
    m = ReadMesh("lug.msh", 3)
    m.KeepVolElems()
    # m.Plot()
    return m


def StructuredMesh(box, dx, typel=3):
    """Build a structured mesh from two points coordinates (box)
    parameters
    ----------
    box: numpy.array
        box = np.array([[xmin, ymin],
                    [xmax, ymax]])    in mesh unit
    dx: numpy or python array
        dx = [dx, dy]: average element size (can be scalar) in mesh unit
    typel: int
        Element types: (see pyxel.ElTypes)
        2: first order triangles (T3)
        3: first order quadrangles (Q4)
        9: second order triangles (T6)
        10: 9-node second order quadrangles (Q9)
        16: 8-node second order quadrangles (Q8)
        4: 4-node first order tetrahedron (Tet4)
        5: 8-node first order hexahedron (Hex8)
    """
    if typel == 2:
        return StructuredMeshT3(box, dx)
    elif typel == 9:
        return StructuredMeshT6(box, dx)
    elif typel == 3:
        return StructuredMeshQ4(box, dx)
    elif typel == 16:
        return StructuredMeshQ8(box, dx)
    elif typel == 10:
        return StructuredMeshQ9(box, dx)
    # elif typel == 4:
    #     return StructuredMeshTet4(box, dx)
    elif typel == 5:
        return StructuredMeshHex8(box, dx)
    elif typel == 17:
        return StructuredMeshHex20(box, dx)


def MeshFrom1DClosedMesh(m, lc=None, order=1, recombine=False):
    if lc is None:
        lc = np.ones(len(m.n)) * m.GetApproxElementSize()
    elif type(lc) == int or type(lc) == float:
        lc = np.ones(len(m.n)) * lc
    gmsh.initialize()
    gmsh.model.add("lug")
    for i in range(len(m.n)):
        gmsh.model.geo.addPoint(m.n[i, 0], m.n[i, 1], 0, lc[i], tag=i)
    for ie in range(len(m.e[1])):
        gmsh.model.geo.addLine(m.e[1][ie, 0], m.e[1][ie, 1], tag=i+ie)
    gmsh.model.geo.addCurveLoop(np.arange(len(m.e[1]))+i, 1)
    gmsh.model.geo.addPlaneSurface([1], 1)
    if recombine:
        # gmsh.model.geo.mesh.setTransfiniteSurface(1, "Left", [])
        gmsh.option.setNumber('Mesh.RecombineAll', 1)
        gmsh.option.setNumber('Mesh.RecombinationAlgorithm', 1)
    if order > 1:
        gmsh.option.setNumber('Mesh.SecondOrderIncomplete', 1)
        gmsh.option.setNumber('Mesh.ElementOrder', order)
    gmsh.option.setNumber('General.Verbosity', 1)
    gmsh.model.geo.synchronize()
    gmsh.model.mesh.generate(2)
    gmsh.write("lug.msh")
    # if '-nopopup' not in sys.argv:
    #     gmsh.fltk.run()
    gmsh.finalize()
    m2d = ReadMesh("lug.msh", 2)
    m2d.Plot()
    return m2d


# %%


def TetraMeshCylinder(x0, y0, z0, R, h, lc):
    """
    Meshes a cylinder geometry with tetrahedral elements

    Parameters
    ----------
    x0,y0,z0 : Center of the base of cylinder  
    R : Radius
    h : height
    lc : mesh length.

    Returns
    -------
    m : mesh object

    """
    print('cylinder')
    gmsh.initialize()
    gmsh.clear()
    gmsh.model.add("cylinder")
    gmsh.model.geo.addPoint(x0, y0, z0, lc, 1)
    gmsh.model.geo.addPoint(x0-R, y0, z0, lc, 2)
    gmsh.model.geo.addPoint(x0+R, y0, z0, lc, 3)
    gmsh.model.geo.addPoint(x0, y0+R, z0, lc, 4)
    gmsh.model.geo.addPoint(x0, y0-R, z0, lc, 5)
    gmsh.model.geo.addCircleArc(2, 1, 4, 1)
    gmsh.model.geo.addCircleArc(4, 1, 3, 2)
    gmsh.model.geo.addCircleArc(3, 1, 5, 3)
    gmsh.model.geo.addCircleArc(5, 1, 2, 4)
    gmsh.model.geo.addCurveLoop([1, 2, 3, 4], 1)
    gmsh.model.geo.addPlaneSurface([1], 1)
    gmsh.model.geo.extrude([(2, 1)], 0, 0, h)
    gmsh.model.geo.synchronize()
    gmsh.option.setNumber('General.Verbosity', 1)
    gmsh.model.mesh.generate(3)

    nums, nodes, e = gmsh.model.mesh.getNodes()
    nodes = nodes.reshape((len(nums), 3))

    nums, els = gmsh.model.mesh.getElementsByType(4)
    nnd = len(els)//len(nums)
    els = els.reshape((len(nums), nnd)) - 1
    e = {4: els.astype(int)}
    m = Mesh(e, nodes, 3)
    # m.CleanMesh()
    return m


def TetraMeshBox(box, dx):
    """
    Meshes a box with tetrahedral elements

    Parameters
    ----------
    box = np.array([[xmin, ymin, zmin],
                    [xmax, ymax, zmax]])    in mesh unit
    dx : average element size (scalar) in mesh unit

    Returns
    -------
    m : mesh object

    """
    gmsh.initialize()
    gmsh.clear()
    gmsh.model.add("box")
    xmin, ymin, zmin = box[0, :]
    xmax, ymax, zmax = box[1, :]
    gmsh.model.geo.addPoint(xmin, ymin, zmin, dx, 1)
    gmsh.model.geo.addPoint(xmax, ymin, zmin, dx, 2)
    gmsh.model.geo.addPoint(xmax, ymax, zmin, dx, 3)
    gmsh.model.geo.addPoint(xmin, ymax, zmin, dx, 4)
    gmsh.model.geo.addLine(1, 2, 1)
    gmsh.model.geo.addLine(2, 3, 2)
    gmsh.model.geo.addLine(3, 4, 3)
    gmsh.model.geo.addLine(4, 1, 4)
    gmsh.model.geo.addCurveLoop([1, 2, 3, 4], 1)
    gmsh.model.geo.addPlaneSurface([1], 1)
    gmsh.model.geo.extrude([(2, 1)], 0, 0, zmax-zmin)
    gmsh.model.geo.synchronize()
    gmsh.option.setNumber('General.Verbosity', 1)
    gmsh.model.mesh.generate(3)
    nums, nodes, e = gmsh.model.mesh.getNodes()
    nodes = nodes.reshape((len(nums), 3))
    nums, els = gmsh.model.mesh.getElementsByType(4)
    nnd = len(els) // len(nums)
    els = els.reshape((len(nums), nnd)) - 1
    e = {4: els.astype(int)}
    m = Mesh(e, nodes, 3)
    return m

#%%

# def Gmsh2Mesh(gmsh, dim=2):
#     """
#     Bulding pyxel mesh from gmsh python object

#     Parameters
#     ----------
#         gmsh : python gmsh object

#     EXAMPLE:
#     ----------
#         import gmsh
#         gmsh.initialize()
#         gmsh.model.add("P")
#         lc = 0.02
#         gmsh.model.geo.addPoint(0, 0.0, 0, 4 * lc, 1)
#         gmsh.model.geo.addPoint(1, 0.0, 0, lc, 2)
#         gmsh.model.geo.addPoint(1, 0.5, 0, lc, 3)
#         gmsh.model.geo.addPoint(0, 0.5, 0, 4 * lc, 4)
#         gmsh.model.geo.addLine(1, 2, 1)
#         gmsh.model.geo.addLine(2, 3, 2)
#         gmsh.model.geo.addLine(3, 4, 3)
#         gmsh.model.geo.addLine(4, 1, 4)
#         gmsh.model.geo.addCurveLoop([1, 2, 3, 4], 1)
#         gmsh.model.geo.addPlaneSurface([1], 1)
#         gmsh.model.geo.synchronize()
#         gmsh.model.mesh.generate(2)
#         m = px.Gmsh2Mesh(gmsh)
#         m.Plot()

#     """
#     # Get direct full node list
#     nums, nodes, e = gmsh.model.mesh.getNodes()
#     nodes = nodes.reshape((len(nums), 3))
#     elems = dict()
#     # Get the Element by type
#     for et in gmsh.model.mesh.getElementTypes():
#         nums, els = gmsh.model.mesh.getElementsByType(et)
#         nnd = len(els) // len(nums)
#         elems[et] = els.reshape((len(nums), nnd)).astype(int) - 1
#     if 15 in elems:
#         del elems[15]  # remove points
#     if 1 in elems:
#         del elems[1]  # remove segments
#     if 8 in elems:
#         del elems[8]  # remove quadratic segments
#     m = Mesh(elems, nodes[:, :dim], dim)
#     return m


#%% Plot a polygon
def PlotLS(ls, coul):
    for i in range(len(ls)):
        plt.plot(ls[i][:, 1], ls[i][:, 0], coul)
    plt.axis('equal')


def RemoveSegments(ls):
    PlotLS(ls, 'k-')
    rep = []
    for i in range(len(ls)):
        if len(ls[i]) < 4:
            rep += [i]
            PlotLS([ls[i]], 'r-')
    return np.delete(ls, rep)

#%% finds the polygons included in other polygons


def lsio(lsi):
    # determine if the line loop truns clockwise or counterclockwise
    c = np.mean(lsi, axis=0)
    sc = 0
    for i in range(len(lsi)-1):
        xn = lsi[[i, i+1], :]
        v = np.diff(xn, axis=0)[0]
        v /= np.linalg.norm(v)
        w = lsi[i, :] - c
        w /= np.linalg.norm(w)
        sc = np.diff(w[::-1]*v)[0]
    return int(sc > 0)


def RayCasting(lsi, p):
    # Ray Casting Algorithm (finds if a point p is in a polygon lsi)
    #      cast a ray from p vertical in the top direction and count the
    #      number of intersected segments.
    nc = 0
    # compute eps with first edge length
    eps = 1e-5 * np.linalg.norm(np.diff(lsi[[0, 1], :], axis=0))
    for i in range(len(lsi)-1):
        xn = lsi[[i, i+1], :]
        w = np.mean(xn, axis=0) - p
        # comparing the x coordinates of the 3 points.
        if w[1] > 0 and abs(xn[0, 0] - p[0]) + abs(xn[1, 0] - p[0]) - abs(xn[1, 0] - xn[0, 0]) < eps:
            # if xp coorespond to the x the first node, dont consider this edge
            if abs(xn[0, 0] - p[0]) > eps:
                nc += 1
    return nc % 2

#%% bulk meshing from polygon

def MeshFromLS(ls, lc, typel='tri'):
    # ls list of segments
    # lc approximate element size
    lsc = np.zeros(len(ls), dtype=int)
    for i in range(len(ls)):
        lsc[i] = lsio(ls[i])
    if np.max(lsc) == 0:
        raise Exception(
            'in pyxel::MeshFromLS, Only counterclockwise list of segments, use flipud')
    white, = np.where(lsc == 0)
    black, = np.where(lsc)
    connect = -np.ones(len(ls))
    for i in range(len(white)):
        for j in range(len(black)):
            if RayCasting(ls[black[j]], ls[white[i]][0, :]):
                connect[white[i]] = black[j]
                break
    gmsh.initialize()
    gmsh.model.add("P")
    nn = 1
    nl = 1
    curvedloop = dict()
    for j in range(len(ls)):
        for i in range(len(ls[j])-1):
            gmsh.model.geo.addPoint(ls[j][i, 0], ls[j][i, 1], 0, lc, nn+i)
        for i in range(len(ls[j])-2):
            gmsh.model.geo.addLine(nn+i, nn+i+1, nl+i)
        gmsh.model.geo.addLine(nn+i+1, nn, nl+i+1)
        curvedloop[j] = nl + np.arange(len(ls[j]) - 1)
        nn += len(ls[j]) - 1
        nl += len(ls[j]) - 1
    ncl = 1
    for i in range(len(ls)):
        if lsc[i]:
            rep, = np.where(connect == i)
            cvl = curvedloop[i]
            for j in rep:
                cvl = np.append(cvl, -curvedloop[j])
            gmsh.model.geo.addCurveLoop(cvl, ncl)
            gmsh.model.geo.addPlaneSurface([ncl], ncl)
            ncl += 1
    gmsh.model.geo.synchronize()
    if typel == 'quad':
        gmsh.option.setNumber("Mesh.RecombineAll", 1)
    # gmsh.fltk.run()
    gmsh.option.setNumber('General.Verbosity', 1)
    gmsh.model.mesh.generate(2)
    gmsh.write('tmp.msh')
    gmsh.finalize()
    m = ReadMesh('tmp.msh')
    print(m.e.keys())
    if 15 in m.e:
        del m.e[15]  # remove vertex
    if 1 in m.e:
        del m.e[1]  # remove segments
    if 8 in m.e:
        del m.e[8]  # remove quadratic segments
    return m


def MeshFromImage(fpix, h, appls=None, typel='tri'):
    """
    Builds a mesh from a graylevel image.

    Parameters
    ----------
    f : NUMPY ARRAY (bool pixel map)
        1: inside, 0: outside the domain to mesh
    h : FLOAT
        approximate finite elements size in pixels
    appls : BOOL
        If 0, does not approximate the list of segments
        Set appls=4 to allow to move each point of the polygon up to
        4 pixels to regularize its shape. Default None (=no approx.)

    Returns
    -------
    m : PYXEL.MESH
        mesh of the domain where f equals to 0

    """
    fpix[:, 0] = 0
    fpix[0, :] = 0
    fpix[:, -1] = 0
    fpix[-1, :] = 0
    pix = ((1-fpix)*255).astype('uint8')
    ls = measure.find_contours(pix, 127, fully_connected='low')
    #PlotLS(ls,'r.-')
    if appls is not None:
        ls = [measure.approximate_polygon(i, appls) for i in ls]
        # PlotLS(ls,'k.-')
    m = MeshFromLS(ls, h, typel)
    # p = np.array([1, 0., 0., np.pi/2])
    # a = Image('')
    # a.pix = f
    m.n = np.c_[m.n[:, 1], -m.n[:, 0]]
    p = np.array([1, 0., 0., 0])
    cam = Camera(p)
    return m, cam

# xi = np.arange(-50,51)
# X, Y = np.meshgrid(xi, xi)
# Z = np.sqrt(X**2+Y**2)
# fpix = Z<20

# plt.imshow(fpix)
